# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# assets: Cache-friendly asset management via content-hash-naming
# Copyright (C) 2009  Red Hat
#
# This file is part of `assets`.
#
# `assets` is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `assets` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `assets`.  If not, see <http://www.gnu.org/licenses/>.

"""
Cache-friendly asset management via content-hash-naming.

`assetslib` is a simple Python library that uses content-hash-naming to do URL
fingerprinting.  For some good background information on URL fingerprinting and
cache optimization, see:

    http://code.google.com/speed/page-speed/docs/caching.html


Introduction
============

The `assetslib.Assets` class does content-hash-naming (aka URL fingerprinting)
so you can use aggressive caching headers without risking that a client might
have an out-of-date version of an asset in its cache.  If the content changes,
the content-hash-name also changes, resulting in a different URL for each
bytewise-unique version of an asset.

Assets are registered using a human-readable key unique to the asset, by which
the content-hash-name of the current *asset version* can be retrieved.  For
example, say we start off with an initial version of an asset like this:

>>> from assetslib import Assets
>>> a = Assets()
>>> a.register(data='I have a colon full of cookie', key='ihacfoc.txt')
'ihacfoc.txt'
>>> a['ihacfoc.txt']  # Lookup the current content-hash-name
'7cec3d7646a7fdb742bd854734d42841d12d3932.txt'

And now say we register a slightly updated version of the same asset:

>>> a.register(data='I have a colon full of cookie!', key='ihacfoc.txt')
'ihacfoc.txt'
>>> a['ihacfoc.txt']
'0d7be6f8dd97ca463cef0727f9ad82003a9da656.txt'

Notice how appending the single ``'!'`` character to the asset content
dramatically changed the content-hash-name.

The extension part of the key is always preserved on the content-hash-name (this
is important so that your web-server can still deliver the correct
*Content-Type* header).  Multi-part extensions can also be used, for example:

>>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
'site.css.gz'
>>> a['site.css.gz']
'3071f65dae784df1fa10361041a30e280c464b01.css.gz'

You can also use a key with no extension, in which case the content-hash-name
will likewise have no extension.  For example:

>>> a.register(data='<The GNU GPLv3>', key='COPYING')
'COPYING'
>>> a['COPYING']
'76f6ffdf4ad8ff7bc7dff5c5f5ea81be5cac1dd4'

At page generation time, your application should typically use
`Assets.asset_url()` to retrieve a full URL rather than just the
content-hash-name.  For example:

>>> a.asset_url('ihacfoc.txt')
'/_/0d7be6f8dd97ca463cef0727f9ad82003a9da656.txt'
>>> a.asset_url('site.css.gz')
'/_/3071f65dae784df1fa10361041a30e280c464b01.css.gz'

When you create an `Assets` instance, the first optional argument is the URL
from which your assets are served, which defaults to ``'/_/'``.  You can mount
your assets at a different URL like this:

>>> a = Assets('/static/')
>>> a.url
'/static/'
>>> a.register(data='I have a colon full of cookie', key='ihacfoc.txt')
'ihacfoc.txt'
>>> a.asset_url('ihacfoc.txt')
'/static/7cec3d7646a7fdb742bd854734d42841d12d3932.txt'

The second optional argument when you create your `Assets` instance is the
directory where your assets are stored.  If you don't specify the directory, a
random secure temporary directory is created for you.  For example:

>>> a = Assets()
>>> a.dir  # doctest:+ELLIPSIS
'/tmp/assets...'

This temporary directory is created simply to make testing and experimentation
easier, and is definitely not what you should use in a production deployment nor
if you're developing under a multi-process server (as each process would have
a different assets directory).  In a production deployment, consider using
something like ``dir='/var/cache/assets'``.

See the `assetslib.Assets` class for its full API.


Configuring your web-server
===========================

Your web-server should return an *Expires* header with a date approximately one
year in the future from the time at which the request is served, which according
to RFC 2616 indicates to the client that the resource never expires:

    To mark a response as "never expires," an origin server sends an Expires
    date approximately one year from the time the response is sent. HTTP/1.1
    servers SHOULD NOT send Expires dates more than one year in the future.

    -- RFC 2616, section 14.21

To correctly configure Apache, you'll need to enable **mod_expires** and then
insert a configuration snippet something like this:

    ::

        Alias /_/ "/var/cache/assets/"
        <Directory "/var/cache/assets/">
            ExpiresActive On
            ExpiresDefault A31536000
        </Directory>

For more information on this Apache configuration, see:

    http://httpd.apache.org/docs/2.2/mod/mod_expires.html

For your convenience, the `assetslib.wsgi` module includes a simple WSGI
application for serving your assets: `assetslib.wsgi.AssetsApp`.  This is
primarily intended for easy in-tree testing and development.  In a production
setting, you're much better off serving your static assets directly from your
web-server.

The `assetslib.wsgi` module also includes the `assetslib.wsgi.never_expires()`
function, which will generate an *Expires* header ``tuple`` ready to be appended
to the WSGI ``response_headers``.  For example, the *Expires* header for a
request made at Unix time ``1234567890`` would look like this:

>>> from assetslib.wsgi import never_expires
>>> never_expires(1234567890)
('Expires', 'Sat, 13 Feb 2010 23:31:30 GMT')


Implementation details
======================

`assetslib` was designed with specific goals in mind:

* The registered assets must work easily from any web-server without the
  web-server having to use (or even know about) `assetslib`

* The assets must be able to be registered and used by multiple processes,
  applications, and servers simultaneously

* The asset key to content-hash-name mapping must use a very simple mechanism so
  equivalent functionality can easily be implemented in other languages, even
  implemented with a few standard *\*nix* commands

* The registered assets and the state of the above mapping needs to be easy to
  inspect in a human-readable way without special software, especially without
  `assetslib` itself

The simple solution to all the above is to use symlinks to store the key to
content-hash-name mapping.  For example:

>>> import os
>>> from os import path
>>> from assetslib import Assets
>>> a = Assets()
>>> sorted(os.listdir(a.dir))
[]
>>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
'site.css.gz'
>>> a['site.css.gz']
'3071f65dae784df1fa10361041a30e280c464b01.css.gz'
>>> sorted(os.listdir(a.dir))
['3071f65dae784df1fa10361041a30e280c464b01.css.gz', 'site.css.gz']

The key is a relative symlink pointing to the content-hash-named file:

>>> os.readlink(path.join(a.dir, 'site.css.gz'))
'3071f65dae784df1fa10361041a30e280c464b01.css.gz'
>>> path.isfile(path.join(a.dir, '3071f65dae784df1fa10361041a30e280c464b01.css.gz'))
True

By default `Assets` uses a SHA1 hash to generate the content-hash-name.  For
example:

>>> try:
...     from hashlib import sha1
... except ImportError:
...     from sha import new as sha1  # Python 2.4 compatibility
...
>>> content = open(a.asset_file('site.css.gz'), 'rb').read()
>>> sha1(content).hexdigest()
'3071f65dae784df1fa10361041a30e280c464b01'
>>> a.asset_file('site.css.gz')  # doctest:+ELLIPSIS
'/tmp/assets.../3071f65dae784df1fa10361041a30e280c464b01.css.gz'


Looking forward
===============

It is the author's humble hope that this simple standard become widely adopted,
that *\*nix* distros use ``'/var/cache/assets'`` as a system-wide assets
location, and that faster websites subsequently flourish.
"""

import os
from os import path
import tempfile
try:
    from hashlib import sha1
except ImportError:
    from sha import new as sha1


__version__ = '0.1.1'


class Assets(object):
    """
    Provides cache-friendly asset management via content-hash-naming.
    """

    def __init__(self, url='/_/', dir=None, prod=True, hash=sha1):
        """
        Initialize.

        :param url: The base URL from which assets are server; the default is
            ``'/_/'``
        :param dir: The directory on the filesytem where assets are stored, e.g.
            ``'/var/cache/assets/'``; if not provided, a secure temporary
            directory is created
        :param prod: If ``True``, operate in production mode; see
            `Assets.asset_url()`
        :param hash: Secure hash used to create content-hash-names; the default
            is ``hashlib.sha1``
        """
        if not dir:
            dir = tempfile.mkdtemp(prefix='assets.')
        if not path.isdir(dir):
            raise ValueError(
                'Assets.dir: not a directory: %r' % (dir,)
            )
        self.url = url.rstrip('/') + '/'
        self.dir = path.abspath(dir)
        self.prod = prod
        self.hash = hash

    def __repr__(self):
        return '%s(%r, %r, %r, %r)' % (self.__class__.__name__,
            self.url, self.dir, self.prod, self.hash.__name__
        )

    def __iter__(self):
        """
        Iterate through the registered asset keys in ascending order.

        For example:

        >>> a = Assets()
        >>> list(a)
        []
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> list(a)
        ['site.css.gz']
        >>> a.register(data='I have a colon full of cookie', key='ihacfoc.txt')
        'ihacfoc.txt'
        >>> list(a)
        ['ihacfoc.txt', 'site.css.gz']
        """
        for key in sorted(os.listdir(self.dir)):
            p = path.join(self.dir, key)
            if path.islink(p):
                yield key

    def __contains__(self, key):
        """
        Return True if there is an asset named *key*.

        For example:

        >>> a = Assets()
        >>> 'site.css.gz' in a
        False
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> 'site.css.gz' in a
        True

        :param key: The key an asset was registered under, e.g.
            ``'site.css.gz'``
        """
        p = path.join(self.dir, key)
        return path.islink(p)

    def __getitem__(self, key):
        """
        Return the content-hash-name of the asset registered under *key*.

        For example:

        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> a['site.css.gz']
        '3071f65dae784df1fa10361041a30e280c464b01.css.gz'

        If no asset is registered under *key*, a ``KeyError`` is raised.  For
        example:

        >>> a['ihacfoc.txt']
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'

        During development it can be helpful to see the human readable key
        instead of the content-hash-name (this is most relevant when using
        `Assets.asset_url()`, which resolves the content-hash-name here).  If
        you create your `Assets` instance with ``prod=False``, the key is
        returned instead of the content-hash-name.  For example:

        >>> dev = Assets(prod=False)
        >>> dev.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> dev['site.css.gz']
        'site.css.gz'

        However, even with ``prod=False``, a ``KeyError`` is still raised if no
        such asset is registered:

        >>> dev['ihacfoc.txt']
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'


        :param key: The key the asset was registered under, e.g.
            ``'site.css.gz'``
        """
        p = path.join(self.dir, key)
        if not path.islink(p):
            raise KeyError(key)
        if self.prod:
            return os.readlink(p)
        return key

    def items(self):
        """
        Iterate through ``(key,chn)`` pairs in ascending order by ``key``.

        For example:

        >>> a = Assets()
        >>> list(a.items())
        []
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> list(a.items())
        [('site.css.gz', '3071f65dae784df1fa10361041a30e280c464b01.css.gz')]

        Note that this method behaves like the Python 3 ``dict.items()`` method,
        which iterates instead of returning a ``list``.
        """
        for key in self:
            yield (key, self[key])

    def register(self, filename=None, fileobj=None, data=None, key=None):
        """
        Register an asset and return its key.

        When registering an asset, you supply the asset data via either the
        *filename*, *fileobj*, or *data* kwarg.  Only one should be supplied and
        if multiple are supplied, the first of the three is used.

        If you provide a *filename* but do not provide a *key*, the ``key`` will
        be set to ``path.basename(filename)``.  On the other hand, if you
        provide the asset data via the *fileobj* or *data* kwargs, you must
        always provide the *key*.

        :param filename: Path of a file to register as an asset, e.g.
            ``'/my/site.css.gz'``
        :param fileobj: An open file descriptor, other ``IOBase`` instance, or
            similar (anything with a ``read()`` method)
        :param data: An ``str`` or ``unicode`` instance providing the data, e.g.
            ``'I have a colon full of cookie'``
        :param key: A human-readable key under which to register the asset, e.g.
            ``'ihacfoc.txt'``
        """
        if filename:
            if not path.isfile(filename):
                raise ValueError(
                    'filename: not a file: %r' % (filename,)
                )
            fileobj = open(filename, 'rb')
            if not key:
                key = path.basename(filename)
        if fileobj:
            if not callable(getattr(fileobj, 'read', None)):
                raise TypeError('fileobj.read() not a callable')
            data = fileobj.read()
        if not data:
            raise TypeError('must provide either filename, fileobj, or data')
        if not key:
            raise TypeError('must provide key if filename=None')

        if key in self:
            self.remove(key)

        name = self.hash(data).hexdigest()
        parts = key.split('.', 1)
        if len(parts) == 2:
            name += ('.' + parts[1])
        open(path.join(self.dir, name), 'wb').write(data)
        os.symlink(name, path.join(self.dir, key))
        return key

    def remove(self, key):
        """
        Remove the asset registered under *key*.

        For example:

        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> 'site.css.gz' in a
        True
        >>> a.remove('site.css.gz')
        >>> 'site.css.gz' in a
        False

        If you try to remove an asset that doesn't exist, a ``KeyError`` will be
        raised.  For example:

        >>> a.remove('site.css.gz')  # Already removed
        Traceback (most recent call last):
          ...
        KeyError: 'site.css.gz'

        :param key: The key of the asset you wish to remove, e.g.
            ``'site.css.gz'``
        """
        name = self[key]
        l = path.join(self.dir, key)
        f = path.join(self.dir, name)
        if path.islink(l):
            os.remove(l)
        if path.isfile(f) and not path.islink(f):
            os.remove(f)

    def asset_url(self, key):
        """
        Return the full URL of the asset registered under *key*.

        For example:

        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> a.asset_url('site.css.gz')
        '/_/3071f65dae784df1fa10361041a30e280c464b01.css.gz'

        If no asset is registered under *key*, a ``KeyError`` is raised.  For
        example:

        >>> a.asset_url('ihacfoc.txt')
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'

        During development it can be helpful to see a URL ending with the human
        readable key instead of the content-hash-name.  If you create your
        `Assets` instance with ``prod=False``, the URL is built from the key
        instead of the content-hash-name.  For example:

        >>> dev = Assets(prod=False)
        >>> dev.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> dev.asset_url('site.css.gz')
        '/_/site.css.gz'

        However, even with ``prod=False``, a ``KeyError`` is still raised if no
        such asset is registered:

        >>> dev.asset_url('ihacfoc.txt')
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'

        :param key: The key of the asset whose URL you wish to retrieve, e.g.
            ``'site.css.gz'``
        """
        return self.url + self[key]

    def asset_file(self, key):
        """
        Return the full file path of the asset registered under *key*.

        For example:

        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> a.asset_file('site.css.gz')  # doctest:+ELLIPSIS
        '/tmp/assets.../3071f65dae784df1fa10361041a30e280c464b01.css.gz'

        If no asset is registered under *key*, a ``KeyError`` is raised.  For
        example:

        >>> a.asset_file('ihacfoc.txt')
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'

        If you create your `Assets` instance with ``prod=False``, the asset
        asset symlink is returned instead of the filename.  For example:

        >>> dev = Assets(prod=False)
        >>> dev.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> dev.asset_file('site.css.gz')  # doctest:+ELLIPSIS
        '/tmp/assets.../site.css.gz'

        However, even with ``prod=False``, a ``KeyError`` is still raised if no
        such asset is registered:

        >>> dev.asset_file('ihacfoc.txt')
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'

        :param key: The key of the asset whose filename you wish to retrieve,
            e.g. ``'site.css.gz'``
        """
        return path.join(self.dir, self[key])

    def asset_symlink(self, key):
        """
        Return the full symlink path of the asset registered under *key*.

        For example:

        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> a.asset_symlink('site.css.gz')  # doctest:+ELLIPSIS
        '/tmp/assets.../site.css.gz'

        `Assets` uses symlinks to maintain the key to content-hash-name mapping,
        as you can see here:

        >>> os.readlink(a.asset_symlink('site.css.gz'))
        '3071f65dae784df1fa10361041a30e280c464b01.css.gz'
        >>> a.asset_file('site.css.gz')  # doctest:+ELLIPSIS
        '/tmp/assets.../3071f65dae784df1fa10361041a30e280c464b01.css.gz'

        If no asset is registered under *key*, a ``KeyError`` is raised.  For
        example:

        >>> a.asset_symlink('ihacfoc.txt')
        Traceback (most recent call last):
          ...
        KeyError: 'ihacfoc.txt'

        Regardless of whether the `Assets` instance was created with
        ``prod=False``, the asset symlink is always returned.

        :param key: The key of the asset whose symlink you wish to retrieve,
            e.g. ``'site.css.gz'``
        """
        p = path.join(self.dir, key)
        if not path.islink(p):
            raise KeyError(key)
        return p
