# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# assets: Cache-friendly asset management via content-hash-naming
# Copyright (C) 2009  Red Hat
#
# This file is part of `assets`.
#
# `assets` is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `assets` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `assets`.  If not, see <http://www.gnu.org/licenses/>.

"""
Simple WSGI application for serving assets.

The `AssetsApp` WSGI application is intended for use in a development server run
from inside your project source tree, for test harnesses, and so on.  In a
production deployment, you're much better off serving your assets directly from
your web-server.

For example, say you create an `Assets` instance like this:

>>> from assetslib import Assets
>>> assets = Assets()
>>> assets.register(data='<My pre-gzipped CSS>', key='site.css.gz')
'site.css.gz'
>>> assets['site.css.gz']
'3071f65dae784df1fa10361041a30e280c464b01.css.gz'

You then provide this `Assets` instance as the first argument when you create an
`AssetsApp` instance:

>>> from assetslib.wsgi import AssetsApp
>>> app = AssetsApp(assets)
>>> app.assets is assets
True
>>> app.url
'/_/'
>>> app.dir  # doctest:+ELLIPSIS
'/tmp/assets...'

And you could simulate a call from a WSGI server like this, requesting an asset
via its content-hash-name:

>>> def start_response(status, response_headers):
...     print status
...     for t in response_headers:
...         print '%s: %s' % t
...
>>> environ = dict(
...     SCRIPT_NAME='/_',
...     PATH_INFO='/3071f65dae784df1fa10361041a30e280c464b01.css.gz',
... )
>>> app.request_time = 1234567890  # Trick to fake the request time
>>> app(environ, start_response)
200 OK
Content-Length: 20
Expires: Sat, 13 Feb 2010 23:31:30 GMT
Content-Type: text/css
Content-Encoding: gzip
['<My pre-gzipped CSS>']

Notice the *Expires* header above.  However, you can also request an assets via
its key, in which case `AssetsApp` will include a *Last-Modified* header
instead.  For example:

>>> environ = dict(SCRIPT_NAME='/_', PATH_INFO='/site.css.gz')
>>> app.mtime = 1234567890  # Trick to fake the modification time
>>> app(environ, start_response)
200 OK
Content-Length: 20
Last-Modified: Fri, 13 Feb 2009 23:31:30 GMT
Content-Type: text/css
Content-Encoding: gzip
['<My pre-gzipped CSS>']


For more information on the Python Web Server Gateway Interface (WSGI) standard,
see:

    http://www.python.org/dev/peps/pep-0333/
"""

from os import path
import mimetypes
from datetime import datetime
import time
from xml.sax.saxutils import escape
from urllib import quote, unquote
from assetslib import Assets

# The mimetypes lazy init generally doesn't play safe with threads, so we
# explicitly init it here:
mimetypes.init()


_one_year =  60 * 60 * 24 * 365
_weekdays = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
_months = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct',
          'Nov', 'Dec')


def rfc_1123(timestamp):
    """
    Convert Unix time *timestamp* to an RFC 1123 formatted string.

    For example:

    >>> rfc_1123(1234567890)
    'Fri, 13 Feb 2009 23:31:30 GMT'

    :param timestamp: Unix time in seconds (``int`` or ``float``)
    """
    dt = datetime.utcfromtimestamp(timestamp)
    return '%s, %02d %s %04d %02d:%02d:%02d GMT' % (
        _weekdays[dt.weekday()],
        dt.day, _months[dt.month - 1], dt.year,
        dt.hour, dt.minute, dt.second
    )


def never_expires(from_time=None):
    """
    Return an *Expires* header tuple with a date one year from now.

    You will typically call this without supplying the ``from_time`` argument,
    in which case the current system time is used.  But for illustration
    purposes, we will provide the ``from_time``:

    >>> never_expires(1234567890)
    ('Expires', 'Sat, 13 Feb 2010 23:31:30 GMT')

    :param from_time: Unix time in seconds (``int`` or ``float``); defaults to
        ``time.time()``
    """
    if from_time is None:
        from_time = time.time()
    return ('Expires', rfc_1123(from_time + _one_year))


_not_found_template = """<html>
<head>
<title>404 Not Found</title>
</head>
<body>
<h1>404 Not Found</h1>
<p>
No such asset: <code>%s</code>
</p>
</body>
</html>"""


_index_start = """<html>
<head>
<title>%(title)s</title>
</head>
<body>
<h1>%(title)s</h1>
<table>
<tr><th>Key</th><th>Content-Hash-Name</th></tr>
"""

_index_row = """<tr>
<td><a href="%s"><code>%s</code></a></td>
<td><a href="%s"><code>%s</code></a></td>
</tr>
"""

_index_end = '</table></body></html>'


class AssetsApp(object):
    """
    WSGI application for serving assets.
    """

    request_time = None
    mtime = None

    def __init__(self, assets, not_found=None, index=None):
        """
        Initialize.

        :param assets: A `assetslib.Assets` instance

        :param not_found: Custom WSGI application for returning a
            ``'404 Not Found'`` error when no such asset exits (defaults to
            `AssetsApp._not_found`)

        :param index: Custom WSGI application to generate a directory-listing
            page when the asset ``''`` is requested (defaults to
            `AssetsApp._index`)
        """
        self.assets = assets
        self.url = assets.url
        self.dir = assets.dir
        if not callable(not_found):
            not_found = self._not_found
        if not callable(index):
            index = self._index
        self.not_found = not_found
        self.index = index

    def __call__(self, environ, start_response):
        key = unquote(environ['PATH_INFO'].lstrip('/'))
        if key == '':
            return self.index(environ, start_response)
        filename = path.join(self.dir, key)
        try:
            data = open(filename, 'rb').read()
        except IOError:
            return self.not_found(environ, start_response)
        response_headers = [
            ('Content-Length', str(len(data))),
        ]
        if path.islink(filename):
            if self.mtime:
                mtime = self.mtime
            else:
                mtime = path.getmtime(filename)
            response_headers.append(
                ('Last-Modified', rfc_1123(mtime))
            )
        else:
            response_headers.append(never_expires(self.request_time))
        (mimetype, encoding) = mimetypes.guess_type(filename)
        if mimetype:
            response_headers.append(
                ('Content-Type', mimetype)
            )
        if encoding:
            response_headers.append(
                ('Content-Encoding', encoding)
            )
        start_response('200 OK', response_headers)
        return [data]

    def _not_found(self, environ, start_response):
        """
        Return a 404 Not Found response.

        For example:

        >>> from assetslib import Assets
        >>> from assetslib.wsgi import AssetsApp
        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> app = AssetsApp(a)
        >>> environ=dict(SCRIPT_NAME='/_', PATH_INFO='/ihacfoc.txt')
        >>> start_response = lambda status, response_headers: None
        >>> print app._not_found(environ, start_response)[0]
        <html>
        <head>
        <title>404 Not Found</title>
        </head>
        <body>
        <h1>404 Not Found</h1>
        <p>
        No such asset: <code>ihacfoc.txt</code>
        </p>
        </body>
        </html>
        """
        path_info = unquote(environ['PATH_INFO'].lstrip('/'))
        response_headers = [('Content-Type', 'text/html; charset=utf-8')]
        start_response('404 Not Found', response_headers)
        return [
            _not_found_template % escape(path_info)
        ]

    def _index(self, environ, start_response):
        """
        Return a diretory-listing-like page.

        For example:

        >>> from assetslib import Assets
        >>> from assetslib.wsgi import AssetsApp
        >>> a = Assets()
        >>> a.register(data='<My pre-gzipped CSS>', key='site.css.gz')
        'site.css.gz'
        >>> a.register(data='I have a colon full of cookie', key='ihacfoc.txt')
        'ihacfoc.txt'
        >>> app = AssetsApp(a)
        >>> environ=dict(SCRIPT_NAME='/_', PATH_INFO='/')
        >>> start_response = lambda status, response_headers: None
        >>> print app._index(environ, start_response)[0]
        <html>
        <head>
        <title>Assets in /_/</title>
        </head>
        <body>
        <h1>Assets in /_/</h1>
        <table>
        <tr><th>Key</th><th>Content-Hash-Name</th></tr>
        <tr>
        <td><a href="ihacfoc.txt"><code>ihacfoc.txt</code></a></td>
        <td><a href="7cec3d7646a7fdb742bd854734d42841d12d3932.txt"><code>7cec3d7646a7fdb742bd854734d42841d12d3932.txt</code></a></td>
        </tr>
        <tr>
        <td><a href="site.css.gz"><code>site.css.gz</code></a></td>
        <td><a href="3071f65dae784df1fa10361041a30e280c464b01.css.gz"><code>3071f65dae784df1fa10361041a30e280c464b01.css.gz</code></a></td>
        </tr>
        </table></body></html>
        """
        script_name = unquote(environ['SCRIPT_NAME'] + environ['PATH_INFO'])
        title = escape('Assets in ' + script_name)
        response_headers = [('Content-Type', 'text/html; charset=utf-8')]
        start_response('200 OK', response_headers)
        return [
            ''.join(self._index_iter(title))
        ]

    def _index_iter(self, title):
        yield _index_start % dict(title=title)
        for (key, chn) in self.assets.items():
            yield _index_row % (
                quote(key), escape(key),
                quote(chn), escape(chn),
            )
        yield _index_end
